"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const functional_1 = require("./functional");
;
;
;
function analyze(processedScenarios, config) {
    const scenarioAnalyses = {};
    for (const scenarioName of Object.keys(processedScenarios)) {
        const processedScenario = processedScenarios[scenarioName];
        const analysis = analyzeScenario(processedScenario, scenarioName, config);
        scenarioAnalyses[scenarioName] = analysis;
    }
    return scenarioAnalyses;
}
exports.analyze = analyze;
/**
 * Process profiler output and check for regressions.
 */
function analyzeScenario(scenario, scenarioName, config) {
    if (scenario.output) {
        let numTicks = getTicks(scenario.output.dataFile);
        if (scenario.baseline && scenario.baseline.output) {
            let numTicksBaseline = getTicks(scenario.baseline.output.dataFile);
            let analysis = functional_1.analyzeFunctions(scenario.baseline.output.dataFile, scenario.output.dataFile);
            if (analysis.isRegression) {
                analysis.regressionFile = `${scenarioName}.regression.txt`;
                fs_1.default.writeFileSync(path_1.default.join(config.outDir, analysis.regressionFile), analysis.summary);
            }
            return {
                numTicks,
                baseline: {
                    numTicks: numTicksBaseline
                },
                regression: analysis
            };
        }
        return { numTicks };
    }
    console.log('analyzeScenario: no output to analyze!');
    return undefined;
}
/**
 * Get ticks from flamegraph file.
 *
 * @param {string} resultsFile
 */
function getTicks(resultsFile) {
    const numTicks = fs_1.default
        .readFileSync(resultsFile, 'utf8')
        .toString()
        .match(/numTicks\s?\=\s?([0-9]+)/);
    if (numTicks && numTicks[1]) {
        return Number.parseInt(numTicks[1]);
    }
    else {
        console.log('Could not read numTicks from ' + resultsFile);
        return -1;
    }
}
