"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const profile_1 = require("./profile");
const process_1 = require("./process");
const analyze_1 = require("./analyze");
;
/**
 * Function that takes a directory path
 * and creates the structure if it doesn't exist
 * @param {string} path directory path
 * @returns {string} the directory path
 */
function resolveDir(dirPath) {
    if (!fs_1.default.existsSync(dirPath)) {
        fs_1.default.mkdirSync(dirPath);
    }
    return path_1.default.resolve(dirPath);
}
/**
 *
 * @param {Scenarios} scenarios Scenarios under test.
 * @param {ScenarioConfig} userConfig Optional user configuration. Paths not provided default to working directory.
 */
function cook(scenarios, userConfig) {
    return __awaiter(this, void 0, void 0, function* () {
        const config = {
            outDir: userConfig && userConfig.outDir ? resolveDir(userConfig.outDir) : process.cwd(),
            tempDir: userConfig && userConfig.tempDir ? resolveDir(userConfig.tempDir) : process.cwd(),
            pageActions: userConfig && userConfig.pageActions,
        };
        const profiles = yield profile_1.profile(scenarios, config);
        const processed = yield process_1.processProfiles(profiles, config);
        const analyses = analyze_1.analyze(processed, config);
        const results = {};
        Object.keys(scenarios).forEach(scenarioName => {
            results[scenarioName] = {
                profile: profiles[scenarioName],
                processed: processed[scenarioName],
                analysis: analyses[scenarioName],
            };
        });
        fs_1.default.writeFileSync(path_1.default.join(config.outDir, 'results.json'), JSON.stringify(results));
        console.log('results: ');
        console.dir(results, { depth: null });
        return results;
    });
}
exports.cook = cook;
;
exports.default = {
    cook
};
