"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
function addMetrics(flamegraph, metrics) {
    const metricsTableHeader = `
    <details>
      <summary>Metrics</summary>
      <table>
        <tr>
          <th>
            <a href="https://github.com/GoogleChrome/puppeteer/blob/master/docs/api.md#pagemetrics">Metric</a>
          </th>
          <th>Value</th>
        </tr>
    `;
    const metricsTableFooter = `
      </table>
    </details>  
  `;
    const metricsTableContents = Object
        .keys(metrics)
        .map(key => {
        return `
          <tr>
            <td>${key}</td>
            <td>${getFormattedValue(metrics[key], key)}</td>
          </tr>
      `;
    })
        .join('');
    const metricsTable = metricsTableHeader + metricsTableContents + metricsTableFooter;
    // Use placeholder in string to add metrics.
    flamegraph = flamegraph
        .split('<!-- /* METRICS_PLACEHOLDER */ -->')
        .join(metricsTable);
    return flamegraph;
}
exports.addMetrics = addMetrics;
;
function getFormattedValue(value, key) {
    switch (key) {
        case 'JSHeapUsedSize':
        case 'JSHeapTotalSize':
            return value.toLocaleString('en') + ' bytes';
        case 'LayoutDuration':
        case 'RecalcStyleDuration':
        case 'ScriptDuration':
        case 'TaskDuration':
            return value.toLocaleString('en', { maximumSignificantDigits: 3 }) + ' s';
        default:
            return value.toString();
    }
}
