"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const tmp = __importStar(require("tmp"));
const profile_1 = require("../profile");
describe('profileUrl', () => {
    const { profileUrl } = profile_1.__unitTestHooks;
    const testUrl = 'testUrl';
    const testMetrics = { metric1: 1, metric2: 2 };
    let testPage;
    const testBrowser = {
        newPage: jest.fn(() => {
            logfile = tmp.fileSync({ dir: outdir.name });
            return Promise.resolve(testPage);
        })
    };
    const testSelector = '#testSelector';
    let outdir;
    let logfile;
    beforeAll(() => {
        outdir = tmp.dirSync({ unsafeCleanup: true });
    });
    afterAll(() => {
        outdir.removeCallback();
    });
    beforeEach(() => {
        testPage = {
            close: jest.fn(() => Promise.resolve()),
            goto: jest.fn(() => {
                return Promise.resolve(null);
            }),
            metrics: jest.fn(() => Promise.resolve(testMetrics)),
            setDefaultTimeout: jest.fn(() => { }),
            waitForSelector: jest.fn(() => Promise.resolve()),
        };
    });
    afterEach(() => {
        jest.clearAllMocks();
    });
    it('performs expected operations', () => __awaiter(this, void 0, void 0, function* () {
        const result = yield profileUrl(testBrowser, testUrl, 'testScenario', outdir.name);
        expect(testPage.goto.mock.calls.length).toEqual(1);
        expect(testPage.goto.mock.calls[0][0]).toEqual(testUrl);
        expect(testPage.close.mock.calls.length).toEqual(1);
        expect(logfile).toBeDefined();
        expect(result.logFile).toEqual(logfile.name);
        expect(result.metrics).toEqual(testMetrics);
    }));
    it('performs expected operations when user defined page actions', () => __awaiter(this, void 0, void 0, function* () {
        const pageActions = (page, options) => __awaiter(this, void 0, void 0, function* () {
            yield page.setDefaultTimeout(0);
            yield page.goto(options.url);
            yield page.waitForSelector(testSelector);
        });
        const result = yield profileUrl(testBrowser, testUrl, 'testScenario', outdir.name, pageActions);
        expect(testPage.setDefaultTimeout.mock.calls.length).toEqual(1);
        expect(testPage.setDefaultTimeout.mock.calls[0][0]).toEqual(0);
        expect(testPage.goto.mock.calls.length).toEqual(1);
        expect(testPage.goto.mock.calls[0][0]).toEqual(testUrl);
        expect(testPage.close.mock.calls.length).toEqual(1);
        expect(testPage.waitForSelector.mock.calls.length).toEqual(1);
        expect(testPage.waitForSelector.mock.calls[0][0]).toEqual(testSelector);
        expect(logfile).toBeDefined();
        expect(result.logFile).toEqual(logfile.name);
        expect(result.metrics).toEqual(testMetrics);
    }));
});
