import puppeteer from 'puppeteer';
import { Browser, Metrics } from 'puppeteer';
import { Scenarios, ScenarioConfig, PageActions } from '../flamegrill';
export interface Profile {
    logFile: string;
    metrics: Metrics;
}
export interface ScenarioProfile extends Profile {
    baseline?: Profile;
}
export interface ScenarioProfiles {
    [scenarioName: string]: ScenarioProfile;
}
export declare type ProfilePage = puppeteer.Page;
declare type Optional<T, K extends keyof T> = Omit<T, K> & Partial<T>;
export declare type ScenarioProfileConfig = Optional<Required<ScenarioConfig>, 'pageActions'>;
/**
 * Profiles scenarios and generates profile (V8) logs.
 *
 * @param config
 * @param scenarios
 */
export declare function profile(scenarios: Scenarios, config: ScenarioProfileConfig): Promise<ScenarioProfiles>;
/**
 * Run profiler against provided URL and return resulting profiler output.
 *
 * @param {*} browser Launched puppeteer instance.
 * @param {string} testUrl Base URL supporting 'scenario' and 'iterations' query parameters.
 * @param {string} profileName Name of scenario that will be used with baseUrl.
 * @param {string} logDir Absolute path to output log profiles.
 * @param {PageActions} pageActions Async opertaion that is executed before taking metrics.
 * @returns {string} Log file path associated with test.
 */
declare function profileUrl(browser: Browser, testUrl: string, profileName: string, logDir: string, pageActions?: PageActions): Promise<Profile>;
export declare const __unitTestHooks: {
    profileUrl: typeof profileUrl;
};
export {};
//# sourceMappingURL=profile.d.ts.map